@@ -40,8 +40,15 @@
     Updated by      Richard Gooch   9-JUN-1999: Added variable expansion
   support.
 
-    Last updated by Richard Gooch   10-JUN-1999: Added "devname" variable and
-  renamed "device" variable name to "devpath".
+    Updated by      Richard Gooch   10-JUN-1999: Added "devname" variable and
+  renamed "device" variable name to "devpath". Fixed bug in argument handling.
+  Added "-d" switch.
+
+    Updated by      Richard Gooch   13-JUN-1999: Added compile-time check of
+  protocol revision.
+
+    Last updated by Richard Gooch   14-JUN-1999: Stat inode instead of using
+  device entry information (which may be out of date).
 
 
 */
@@ -67,9 +74,18 @@
 #include <karma.h>
 
 
-#define VERSION "1.2.1"
+#define VERSION "1.2.2"
 #define CONFIG_FILE "/etc/devfsd.conf"
-#define MAX_ARGS 6
+#define MAX_ARGS (6 + 1)
+
+/*  Update only after changing code to reflect new protocol  */
+#define DEVFSD_PROTOCOL_REVISION_DAEMON  3
+
+/*  Compile-time check  */
+#if DEVFSD_PROTOCOL_REVISION_KERNEL != DEVFSD_PROTOCOL_REVISION_DAEMON
+#error protocol version mismatch. Update your kernel headers
+#endif
+
 
 #define ACTION_PERMISSIONS              0x0000
 #define ACTION_EXECUTE_ON_REGISTER      0x0100
@@ -78,6 +94,12 @@
 #define ACTION_EXECUTE_ON_CLOSE         0x0400
 #define ACTION_EXECUTE_MASK             0xff00
 
+#define SYSLOG(level, format, args...) \
+    {if (trace_level > 0) fprintf (stderr, format, ## args); \
+     else syslog (level, format, ##args); \
+    }
+
+
 struct permissions_type
 {
     mode_t mode;
@@ -114,6 +136,7 @@ EXTERN_FUNCTION (flag st_expr_expand,
 
 
 /*  Private functions  */
+static void setup_initial_entries (CONST char *mount_point);
 static void read_config ();
 static void do_servicing (int fd);
 static void service_name (const struct devfsd_notify_struct *info);
@@ -122,7 +145,7 @@ static void update_permissions (const struct devfsd_notify_struct *info,
 static void execute (const struct devfsd_notify_struct *info,
 		     const struct config_entry *entry);
 static void free_config ();
-static void do_test (int fd);
+static void do_debug (int fd);
 static uid_t get_uid (const char *string);
 static gid_t get_gid (const char *string);
 static mode_t get_mode (const char *string);
@@ -134,28 +157,53 @@ static CONST char *get_variable (CONST char *variable, void *info);
 static struct config_entry *first_config = NULL;
 static struct config_entry *last_config = NULL;
 static const char *mount_point = NULL;
-
+static char trace_level = 0;
+static struct initial_symlink_struct
+{
+    char *dest;
+    char *name;
+} initial_symlinks[] =
+{
+    {"/proc/self/fd", "fd"},
+    {"fd/0", "stdin"},
+    {"fd/1", "stdout"},
+    {"fd/2", "stderr"},
+    {NULL, NULL},
+};
+    
 
 /*  Public functions follow  */
 
 int main (int argc, char **argv)
 {
     flag print_version = FALSE;
-    flag test = FALSE;
-    int fd, proto_rev;
+    flag debug = FALSE;
+    flag debug_devfs = FALSE;
+    int fd, proto_rev, count;
+    unsigned int devfs_debug_mask;
     struct sigaction new_action;
+    struct stat statbuf;
     char filename[STRING_LENGTH];
-    static char usage[] = "devfsd mntpnt [-vt]";
+    static char usage[] = "devfsd mntpnt [-v] [-d] [-t num] [-D mask]";
 
-    if ( (argc < 2) || (argc > 3) )
+    if (argc < 2)
     {
 	fprintf (stderr, "Usage:\t%s\n", usage);
 	exit (1);
     }
-    if (argc == 3)
+    for (count = 2; count < argc; ++count)
     {
-	if (strcmp (argv[2], "-v") == 0) print_version = TRUE;
-	else if (strcmp (argv[2], "-t") == 0) test = TRUE;
+	if (strcmp (argv[count], "-v") == 0) print_version = TRUE;
+	else if (strcmp (argv[count], "-d") == 0) debug = TRUE;
+	else if ( (strcmp (argv[count], "-t") == 0) && (++count < argc) )
+	{
+	    trace_level = atoi (argv[count]);
+	}
+	else if ( (strcmp (argv[count], "-D") == 0) && (++count < argc) )
+	{
+	    devfs_debug_mask = strtol (argv[count], NULL, 0);
+	    debug_devfs = TRUE;
+	}
 	else
 	{
 	    fprintf (stderr, "Usage:\t%s\n", usage);
@@ -165,18 +213,6 @@ int main (int argc, char **argv)
     mount_point = argv[1];
     strcpy (filename, argv[1]);
     strcat (filename, "/.devfsd");
-    if (print_version)
-    {
-	fprintf (stderr, "devfsd: Linux device filesystem daemon v%s\n",
-		 VERSION);
-	fprintf (stderr,
-		 "(C) 1998-1999  Richard Gooch <rgooch@atnf.csiro.au>\n\n");
-    }
-    if (print_version || test)
-    {
-	fprintf (stderr, "Daemon protocol revision:\t%d\n",
-		 DEVFSD_PROTOCOL_REVISION);
-    }
     if ( ( fd = open (filename, O_RDONLY, 0) ) < 0 )
     {
 	fprintf (stderr, "Error opening file: \"%s\"\t%s\n",
@@ -188,48 +224,68 @@ int main (int argc, char **argv)
 	fprintf (stderr, "Error getting protocol revision\t%s\n", ERRSTRING);
 	exit (1);
     }
-    if ( print_version || test || (DEVFSD_PROTOCOL_REVISION != proto_rev) )
+    if ( print_version || (trace_level > 0) || debug ||
+	 (DEVFSD_PROTOCOL_REVISION_DAEMON != proto_rev) )
     {
-	if (!test) close (fd);
+	fprintf (stderr, "devfsd: Linux device filesystem daemon v%s\n",
+		 VERSION);
+	fprintf (stderr,
+		 "(C) 1998-1999  Richard Gooch <rgooch@atnf.csiro.au>\n\n");
 	fprintf (stderr, "Daemon protocol revision:\t%d\n",
-		 DEVFSD_PROTOCOL_REVISION);
+		 DEVFSD_PROTOCOL_REVISION_DAEMON);
 	fprintf (stderr, "Kernel-side protocol revision:\t%d\n", proto_rev);
-	if (DEVFSD_PROTOCOL_REVISION != proto_rev)
+	if (DEVFSD_PROTOCOL_REVISION_DAEMON != proto_rev)
 	{
 	    fprintf (stderr, "Protocol mismatch!\n");
 	    exit (1);
 	}
-	if (!test) exit (0);
     }
-    if (test) do_test (fd);
+    if (print_version) exit (0);
+    if (debug_devfs)
+    {
+	if (ioctl (fd, DEVFSDIOC_SET_DEBUG_MASK, &devfs_debug_mask) != 0)
+	{
+	    fprintf (stderr, "Error setting debug mask\t%s\n", ERRSTRING);
+	    exit (1);
+	}
+	exit (0);
+    }
+    if (debug) do_debug (fd);
+    setup_initial_entries (mount_point);
     sigemptyset (&new_action.sa_mask);
     new_action.sa_flags = 0;
-    /*  Set up control_c handler  */
+    /*  Set up SIGHUP handler  */
     new_action.sa_handler = sig_handler;
     if (sigaction (SIGHUP, &new_action, (struct sigaction *) NULL) != 0)
     {
 	fprintf (stderr, "Error setting SIGHUP handler\t%s\n", ERRSTRING);
 	exit (1);
     }
-    switch ( fork () )
+    if (stat (CONFIG_FILE, &statbuf) != 0) exit (0);
+    if (statbuf.st_size == 0) exit (0);
+    fprintf (stderr, "Started device management daemon for %s\n", mount_point);
+    if (trace_level < 1)
     {
-      case 0:
-	/*  Child  */
-	break;
-      case -1:
-	/*  Error  */
-	perror ("devfsd");
-	exit (2);
-	/*break;*/
-      default:
-	/*  Parent  */
-	exit (0);
-	/*break;*/
+	switch ( fork () )
+	{
+	  case 0:
+	    /*  Child  */
+	    break;
+	  case -1:
+	    /*  Error  */
+	    perror ("devfsd");
+	    exit (2);
+	    /*break;*/
+	  default:
+	    /*  Parent  */
+	    exit (0);
+	    /*break;*/
+	}
+	openlog ("devfsd", LOG_PID, LOG_DAEMON);
+	close (0);
+	close (1);
+	close (2);
     }
-    openlog ("devfsd", LOG_PID, LOG_DAEMON);
-    close (0);
-    close (1);
-    close (2);
     while (TRUE)
     {
 	read_config ();
@@ -241,6 +297,21 @@ int main (int argc, char **argv)
 
 
 /*  Private functions follow  */
+
+static void setup_initial_entries (CONST char *mount_point)
+{
+    struct initial_symlink_struct *curr;
+    char name[STRING_LENGTH];
+
+    for (curr = initial_symlinks; curr->dest != NULL; ++curr)
+    {
+	strcpy (name, mount_point);
+	strcat (name, "/");
+	strcat (name, curr->name);
+	symlink (curr->dest, name);
+    }
+}   /*  End Function setup_initial_entries  */
+
 static void read_config ()
 /*  [SUMMARY] Read the configuration file.
     [RETURNS] Nothing.
@@ -252,8 +323,8 @@ static void read_config ()
 
     if ( ( fp = fopen (CONFIG_FILE, "r") ) == NULL )
     {
-	syslog (LOG_ERR, "error opening file: \"%s\"\t%s\n",
-		 CONFIG_FILE, ERRSTRING);
+	SYSLOG (LOG_ERR, "error opening file: \"%s\"\t%s\n",
+		CONFIG_FILE, ERRSTRING);
 	exit (1);
     }
     while (fgets (buf, STRING_LENGTH, fp) != NULL)
@@ -269,19 +340,19 @@ static void read_config ()
 	if (line[0] == '\0') continue;
 	if (line[0] == '#') continue;
 	for (count = 0; count < MAX_ARGS; ++count) p[count][0] = '\0';
-	num_args = sscanf (line, "%s %s %s %s %s %s %s %s",
+	num_args = sscanf (line, "%s %s %s %s %s %s %s %s %s",
 			   key, name, p[0], p[1], p[2], p[3],
-			   p[4], p[5]);
+			   p[4], p[5], p[6]);
 	if (num_args < 2)
 	{
-	    syslog (LOG_ERR, "bad config line: \"%s\"\n", line);
-	    syslog (LOG_ERR, "exiting\n");
+	    SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
+	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
 	if ( ( new = malloc (sizeof *new) ) == NULL )
 	{
-	    syslog (LOG_ERR, "error allocating\n");
-	    syslog (LOG_ERR, "exiting\n");
+	    SYSLOG (LOG_ERR, "error allocating\n");
+	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
 	memset (new, 0, sizeof *new);
@@ -297,15 +368,15 @@ static void read_config ()
 	    new->action = ACTION_EXECUTE_ON_CLOSE;
 	else
 	{
-	    syslog (LOG_ERR, "bad config line: \"%s\"\n", line);
-	    syslog (LOG_ERR, "exiting\n");
+	    SYSLOG (LOG_ERR, "bad config line: \"%s\"\n", line);
+	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
 	if ( ( err = regcomp (&new->preg, name, 0) ) != 0 )
 	{
 	    regerror (err, &new->preg, tmp, STRING_LENGTH);
-	    syslog (LOG_ERR, "error compiling regexp: \"%s\"\t%s\n", name,tmp);
-	    syslog (LOG_ERR, "exiting\n");
+	    SYSLOG (LOG_ERR, "error compiling regexp: \"%s\"\t%s\n", name,tmp);
+	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
 	switch (new->action)
@@ -314,8 +385,8 @@ static void read_config ()
 	    /*  Get user and group  */
 	    if ( ( ptr = strchr (p[0], '.') ) == NULL )
 	    {
-		syslog (LOG_ERR, "missing '.' character in: \"%s\"\n", p[0]);
-		syslog (LOG_ERR, "exiting\n");
+		SYSLOG (LOG_ERR, "missing '.' character in: \"%s\"\n", p[0]);
+		SYSLOG (LOG_ERR, "exiting\n");
 		exit (1);
 	    }
 	    *ptr++ = '\0';
@@ -334,8 +405,8 @@ static void read_config ()
 		err = strlen (p[count]) + 1;
 		if ( ( new->u.execute.argv[count] = malloc (err) ) == NULL )
 		{
-		    syslog (LOG_ERR, "error allocating\n");
-		    syslog (LOG_ERR, "exiting\n");
+		    SYSLOG (LOG_ERR, "error allocating\n");
+		    SYSLOG (LOG_ERR, "exiting\n");
 		    exit (1);
 		}
 		strcpy (new->u.execute.argv[count], p[count]);
@@ -349,7 +420,7 @@ static void read_config ()
 	last_config = new;
     }
     fclose (fp);
-    syslog (LOG_INFO, "read config file: \"%s\"\n", CONFIG_FILE);
+    SYSLOG (LOG_INFO, "read config file: \"%s\"\n", CONFIG_FILE);
 }   /*  End Function read_config   */
 
 static void do_servicing (int fd)
@@ -365,13 +436,13 @@ static void do_servicing (int fd)
     {
 	if (bytes < 1)
 	{
-	    syslog (LOG_ERR, "devfs closed file!\n");
+	    SYSLOG (LOG_ERR, "devfs closed file!\n");
 	    exit (1);
 	}
 	service_name (&info);
     }
     if (errno == EINTR) return;
-    syslog (LOG_ERR, "Error reading control file\t%s\n", ERRSTRING);
+    SYSLOG (LOG_ERR, "Error reading control file\t%s\n", ERRSTRING);
     exit (1);
 }   /*  End Function do_servicing  */
 
@@ -426,7 +497,7 @@ static void service_name (const struct devfsd_notify_struct *info)
 	    }
 	    break;
 	  default:
-	    syslog (LOG_ERR, "Unknown event type: %u\n", info->type);
+	    SYSLOG (LOG_ERR, "Unknown event type: %u\n", info->type);
 	    exit (1);
 	    /*break;*/
 	}
@@ -442,31 +513,40 @@ static void update_permissions (const struct devfsd_notify_struct *info,
 */
 {
     mode_t new_mode;
+    struct stat statbuf;
     char path[STRING_LENGTH];
 
     strcpy (path, mount_point);
     strcat (path, "/");
     strcat (path, info->devname);
-    new_mode = (info->mode & S_IFMT) | (entry->u.permissions.mode & ~S_IFMT);
-    if (new_mode != info->mode)
+    if (lstat (path, &statbuf) != 0)
+    {
+	SYSLOG (LOG_ERR, "error stat(2)ing: \"%s\"\t%s\n",
+		path, ERRSTRING);
+	SYSLOG (LOG_ERR, "exiting\n");
+	exit (1);
+    }
+    new_mode = (statbuf.st_mode & S_IFMT) |
+	(entry->u.permissions.mode & ~S_IFMT);
+    if (new_mode != statbuf.st_mode)
     {
 	if (chmod (path, new_mode) != 0)
 	{
-	    syslog (LOG_ERR, "error changing mode for: \"%s\"\t%s\n",
+	    SYSLOG (LOG_ERR, "error changing mode for: \"%s\"\t%s\n",
 		    path, ERRSTRING);
-	    syslog (LOG_ERR, "exiting\n");
+	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
     }
-    if ( (entry->u.permissions.uid != info->uid) ||
-	 (entry->u.permissions.gid != info->gid) )
+    if ( (entry->u.permissions.uid != statbuf.st_uid) ||
+	 (entry->u.permissions.gid != statbuf.st_gid) )
     {
 	if (chown (path, entry->u.permissions.uid, entry->u.permissions.gid)
 	    != 0)
 	{
-	    syslog (LOG_ERR, "error changing mode for: \"%s\"\t%s\n",
+	    SYSLOG (LOG_ERR, "error changing mode for: \"%s\"\t%s\n",
 		    path, ERRSTRING);
-	    syslog (LOG_ERR, "exiting\n");
+	    SYSLOG (LOG_ERR, "exiting\n");
 	    exit (1);
 	}
     }
@@ -481,26 +561,21 @@ static void execute (const struct devfsd_notify_struct *info,
 */
 {
     unsigned int count;
-    char *argv[MAX_ARGS];
-    char largv[MAX_ARGS][STRING_LENGTH];
+    char *argv[MAX_ARGS + 1];
+    char largv[MAX_ARGS + 1][STRING_LENGTH];
     char devname[STRING_LENGTH];
 
     strcpy (devname, mount_point);
     strcat (devname, "/");
     strcat (devname, info->devname);
-    argv[MAX_ARGS] = NULL;
-    for (count = 0; count < MAX_ARGS; ++count)
+    for (count = 0; entry->u.execute.argv[count] != NULL; ++count)
     {
-	if (entry->u.execute.argv[count] == NULL)
-	{
-	    argv[count] = NULL;
-	    break;
-	}
 	st_expr_expand (largv[count], STRING_LENGTH,
 			entry->u.execute.argv[count], get_variable, devname,
 			NULL);
 	argv[count] = largv[count];
     }
+    argv[count] = NULL;
     switch ( fork () )
     {
       case 0:
@@ -508,7 +583,7 @@ static void execute (const struct devfsd_notify_struct *info,
 	break;
       case -1:
 	/*  Error  */
-	syslog (LOG_ERR, "error forking\t%s\n", ERRSTRING);
+	SYSLOG (LOG_ERR, "error forking\t%s\n", ERRSTRING);
 	exit (2);
 	/*break;*/
       default:
@@ -518,7 +593,7 @@ static void execute (const struct devfsd_notify_struct *info,
 	/*break;*/
     }
     execvp (argv[0], argv);
-    syslog (LOG_ERR, "error execing: \"%s\"\t%s\n", argv[0], ERRSTRING);
+    SYSLOG (LOG_ERR, "error execing: \"%s\"\t%s\n", argv[0], ERRSTRING);
 }   /*  End Function execute  */
 
 static void free_config ()
@@ -548,8 +623,8 @@ static void free_config ()
     last_config = NULL;
 }   /*  End Function free_config  */
 
-static void do_test (int fd)
-/*  [SUMMARY] Test the devfs protocol.
+static void do_debug (int fd)
+/*  [SUMMARY] Debug the devfs protocol.
     <fd> The open control file.
     [RETURNS] Nothing.
 */
@@ -558,7 +633,7 @@ static void do_test (int fd)
     struct devfsd_notify_struct info;
     char type[STRING_LENGTH];
 
-    fprintf (stderr, "Running in test mode: no actions will be taken...\n");
+    fprintf (stderr, "Running in debug mode: no actions will be taken...\n");
     while ( ( bytes = read (fd, (char *) &info, sizeof info) ) >= 0 )
     {
 	if (bytes < 1)
@@ -591,7 +666,7 @@ static void do_test (int fd)
     }
     fprintf (stderr, "Error reading control file\t%s\n", ERRSTRING);
     exit (1);
-}   /*  End Function do_test  */
+}   /*  End Function do_debug  */
 
 static uid_t get_uid (const char *string)
 /*  [SUMMARY] Convert a string to a UID value.
@@ -604,8 +679,8 @@ static uid_t get_uid (const char *string)
     if ( isdigit (string[0]) ) return atoi (string);
     if ( ( pw_ent = getpwnam (string) ) == NULL )
     {
-	syslog (LOG_ERR, "unknown user: \"%s\"\n", string);
-	syslog (LOG_ERR, "exiting\n");
+	SYSLOG (LOG_ERR, "unknown user: \"%s\"\n", string);
+	SYSLOG (LOG_ERR, "exiting\n");
 	exit (1);
     }
     return (pw_ent->pw_uid);
@@ -622,8 +697,8 @@ static gid_t get_gid (const char *string)
     if ( isdigit (string[0]) ) return atoi (string);
     if ( ( grp_ent = getgrnam (string) ) == NULL )
     {
-	syslog (LOG_ERR, "unknown group: \"%s\"\n", string);
-	syslog (LOG_ERR, "exiting\n");
+	SYSLOG (LOG_ERR, "unknown group: \"%s\"\n", string);
+	SYSLOG (LOG_ERR, "exiting\n");
 	exit (1);
     }
     return (grp_ent->gr_gid);
@@ -640,8 +715,8 @@ static mode_t get_mode (const char *string)
     if ( isdigit (string[0]) ) return strtoul (string, NULL, 8);
     if (strlen (string) != 9)
     {
-	syslog (LOG_ERR, "bad symbolic mode: \"%s\"\n", string);
-	syslog (LOG_ERR, "exiting\n");
+	SYSLOG (LOG_ERR, "bad symbolic mode: \"%s\"\n", string);
+	SYSLOG (LOG_ERR, "exiting\n");
 	exit (1);
     }
     mode = 0;
@@ -659,7 +734,7 @@ static mode_t get_mode (const char *string)
 
 static void sig_handler (int sig)
 {
-    syslog (LOG_INFO, "Caught SIGHUP\n");
+    SYSLOG (LOG_INFO, "Caught SIGHUP\n");
 }   /*  End Function sig_handler  */
 
 static CONST char *get_variable (CONST char *variable, void *info)
